import json
from collections import OrderedDict


class FintectureResponseBase(object):
    def __init__(self, url, code, headers):
        self.url = url
        self.code = code
        self.headers = headers

    @property
    def x_request_id(self):
        try:
            return self.headers["x-request-id"]
        except KeyError:
            return None


class FintectureResponse(FintectureResponseBase):
    def __init__(self, url, body, code, headers):
        FintectureResponseBase.__init__(self, url, code, headers)
        self.body = body
        try:
            self.data = json.loads(body, object_pairs_hook=OrderedDict)
        except Exception:
            # keep Fintecture error format
            self.data = {
                "status": code,
                "code": "internal_invalid_json",
                "errors": [
                    {
                        "code": "internal_invalid_json",
                        "title": "Invalid JSON",
                        "message": "Received JSON content has an invalid format",
                    }
                ],
            }



# File generated from our OpenAPI spec

from . import api_resources


OBJECT_CLASSES = {
    # data structures
    api_resources.ListObject.OBJECT_NAME: api_resources.ListObject,
    api_resources.SearchResultObject.OBJECT_NAME: api_resources.SearchResultObject,
    # business objects
    api_resources.AIS.OBJECT_NAME: api_resources.AIS,
    api_resources.Account.OBJECT_NAME: api_resources.Account,
    api_resources.AccountHolder.OBJECT_NAME: api_resources.AccountHolder,
    api_resources.Application.OBJECT_NAME: api_resources.Application,
    api_resources.Customer.OBJECT_NAME: api_resources.Customer,
    api_resources.EphemeralKey.OBJECT_NAME: api_resources.EphemeralKey,
    api_resources.Payment.OBJECT_NAME: api_resources.Payment,
    api_resources.Payout.OBJECT_NAME: api_resources.Payout,
    api_resources.PIS.OBJECT_NAME: api_resources.PIS,
    api_resources.Provider.OBJECT_NAME: api_resources.Provider,
    api_resources.Refund.OBJECT_NAME: api_resources.Refund,
    api_resources.TestAccount.OBJECT_NAME: api_resources.TestAccount,
}



# File generated from our OpenAPI spec

from .abstract import ListableAPIResource
from .abstract import SearchableAPIResource


class TestAccount(
    ListableAPIResource,
    SearchableAPIResource,
):
    OBJECT_NAME = "testaccount"

    @classmethod
    def search(cls, *args, **kwargs):
        return cls._search(search_url="/res/v1/testaccounts", *args, **kwargs)



from .api_resource import APIResource


class SearchableAPIResource(APIResource):
    @classmethod
    def _search(
        cls, search_url, app_id=None, fintecture_version=None, **params
    ):
        return cls._static_request(
            "get",
            search_url,
            app_id=app_id,
            fintecture_version=fintecture_version,
            params=params,
        )



from .api_resource import APIResource


class ListableAPIResource(APIResource):
    @classmethod
    def auto_paging_iter(cls, *args, **params):
        return cls.list(*args, **params).auto_paging_iter()

    @classmethod
    def list(cls, app_id=None, fintecture_version=None, **params):
        return cls._static_request(
            "get",
            cls.class_url(),
            app_id=app_id,
            fintecture_version=fintecture_version,
            params=params,
        )



# File generated from our OpenAPI spec

from .. import util
from .abstract import APIResourceTestHelpers
from .abstract import CreateableAPIResource
from .abstract import test_helpers


@test_helpers
class Refund(
    CreateableAPIResource,
):
    OBJECT_NAME = "refund"

    @classmethod
    def class_url(cls):
        return "/pis/v2/refund"

    @classmethod
    def list_for_payment(cls, session_id, **params):
        return cls._static_request(
            "get",
            "/pis/v2/payments/{session_id}/refunds".format(
                session_id=util.sanitize_id(session_id)
            ),
            params=params,
        )

    class TestHelpers(APIResourceTestHelpers):
        @classmethod
        def _cls_expire(
            cls, refund, app_id=None, fintecture_version=None, **params
        ):
            return cls._static_request(
                "post",
                "/v1/test_helpers/refunds/{refund}/expire".format(
                    refund=util.sanitize_id(refund)
                ),
                app_id=app_id,
                fintecture_version=fintecture_version,
                params=params,
            )

        @util.class_method_variant("_cls_expire")
        def expire(self, **params):
            return self.resource._request(
                "post",
                "/v1/test_helpers/refunds/{refund}/expire".format(
                    refund=util.sanitize_id(self.resource.get("id"))
                ),
                params=params,
            )



# -*- coding: utf-8 -*-

# Model types for AIS authentication
DECOUPLED_MODEL_TYPE = "decoupled"
REDIRECT_MODEL_TYPE = "redirect"

# Headers for signature verification
SIGNED_HEADER_PARAMETER_LIST = [
    "(request-target)",
    "Date",
    "Digest",
    "X-Request-ID",
]

# API base URLs - tenant specific
PRODUCTION_API_BASE = "https://api.virementmaitrise.societegenerale.eu"
SANDBOX_API_BASE = "https://api.sandbox.virementmaitrise.societegenerale.eu"



# flake8: noqa

from .error_object import ErrorObject
from .list_object import ListObject
from .search_result_object import SearchResultObject

from . import test_helpers

from .ais import AIS
from .application import Application
from .account import Account
from .account_holder import AccountHolder
from .customer import Customer

from .ephemeral_key import EphemeralKey
from .payout import Payout
from .pis import PIS
from .payment import Payment
from .provider import Provider
from .refund import Refund
from .test_account import TestAccount



ENVIRONMENT_TEST = "test"
ENVIRONMENT_SANDBOX = "sandbox"
ENVIRONMENT_PRODUCTION = "production"

DEFAULT_ENV = ENVIRONMENT_SANDBOX

AVAILABLE_ENVS = [
    ENVIRONMENT_TEST,
    ENVIRONMENT_SANDBOX,
    ENVIRONMENT_PRODUCTION,
]



from ... import util
from .api_resource import APIResource
from urllib.parse import quote_plus


class DeletableAPIResource(APIResource):
    @classmethod
    def _cls_delete(cls, sid, **params):
        url = "%s/%s" % (cls.class_url(), quote_plus(util.utf8(sid)))
        return cls._static_request("delete", url, params=params)

    @util.class_method_variant("_cls_delete")
    def delete(self, **params):
        return self._request_and_refresh(
            "delete", self.instance_url(), params=params
        )



import base64

# Get reference to SDK module (works for any package name)
import sys

sdk = sys.modules[__name__.split(".")[0]]
from . import api_requestor, error


class OAuth(object):

    @staticmethod
    def _set_app_params(params):
        if "app_id" in params and "app_secret" in params:
            return

        # Get values from SDK module
        app_id = sdk.app_id
        app_secret = sdk.app_secret

        no_params = False
        if app_id:
            params["app_id"] = app_id
        else:
            no_params = True

        if app_secret:
            params["app_secret"] = app_secret
        else:
            no_params = True

        if no_params:
            raise error.AuthenticationError(
                "No app_id and/or app_secret provided. Set them using: "
                "<package>.app_id = <APP-ID> and <package>.app_secret = <APP-SECRET>"
            )

    @staticmethod
    def ais_token(app_id=None, app_secret=None, **params):
        if app_id:
            params["app_id"] = app_id
        if app_secret:
            params["app_secret"] = app_secret
        OAuth._set_app_params(params)
        return sdk.AIS.oauth(params)

    @staticmethod
    def pis_token(app_id=None, app_secret=None, **params):
        if app_id:
            params["app_id"] = app_id
        if app_secret:
            params["app_secret"] = app_secret
        OAuth._set_app_params(params)
        return sdk.PIS.oauth(app_id, app_secret, **params)

    @staticmethod
    def refresh_token(app_id=None, app_secret=None, **params):
        if not params.get("refresh_token", False):
            raise error.InvalidRequestError(
                message="refresh_token parameter is required for refresh token",
                param="refresh_token",
            )

        if app_id:
            params["app_id"] = app_id
        if app_secret:
            params["app_secret"] = app_secret

        OAuth._set_app_params(params)

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": "Basic {}".format(
                base64.b64encode(
                    "{}:{}".format(
                        params["app_id"], params["app_secret"]
                    ).encode("utf-8")
                ).decode("utf-8")
            ),
        }

        params.update(
            {
                "grant_type": "refresh_token",
            }
        )
        requestor = api_requestor.APIRequestor(
            app_id=params["app_id"],
            app_secret=params["app_secret"],
        )

        response, _ = requestor.request(
            "post", "/oauth/refreshtoken", params, headers
        )
        return response.data



# File generated from our OpenAPI spec

from .. import api_requestor
from .. import util
from .abstract import DeletableAPIResource


class EphemeralKey(DeletableAPIResource):
    OBJECT_NAME = "ephemeral_key"

    @classmethod
    def create(cls, app_id=None, fintecture_version=None, **params):
        if fintecture_version is None:
            raise ValueError(
                "fintecture_version must be specified to create an ephemeral "
                "key"
            )

        requestor = api_requestor.APIRequestor(
            app_id, api_version=fintecture_version
        )

        url = cls.class_url()
        response, my_app_id = requestor.request("post", url, params, {})
        return util.convert_to_fintecture_object(
            response, my_app_id, fintecture_version
        )



import json
from collections import OrderedDict

# Get reference to SDK module (works for any package name)
import sys

sdk = sys.modules[__name__.split(".")[0]]


class FintectureError(Exception):
    def __init__(
        self,
        message=None,
        url=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
    ):
        super(FintectureError, self).__init__(message)

        if http_body and hasattr(http_body, "decode"):
            try:
                http_body = http_body.decode("utf-8")
            except BaseException:
                http_body = (
                    "<Could not decode body as utf-8. "
                    "Please report to contact@fintecture.com>"
                )

            if not json_body:
                try:
                    json_body = json.loads(
                        http_body, object_pairs_hook=OrderedDict
                    )
                except Exception:
                    http_body = (
                        "<Could not parse http body as a JSON format. "
                        "Please report to contact@fintecture.com>"
                    )

        self._message = message
        self.url = url
        self.http_body = http_body
        self.http_status = http_status
        self.json_body = json_body
        self.log_id = json_body.get("log_id", None) if json_body else None
        self.errors = json_body.get("errors", None) if json_body else None
        self.headers = headers or {}
        self.code = code
        self.x_request_id = self.headers.get("x-request-id", None)
        self.error = self.construct_error_object()

    def __str__(self):
        msg = self._message or "<empty message>"
        if self.x_request_id is not None:
            if self.log_id is not None:
                return "Request {0} - Log ID {1}: {2}".format(
                    self.x_request_id, self.log_id, msg
                )
            else:
                return "Request {0}: {1}".format(self.x_request_id, msg)
        else:
            if self.log_id is not None:
                return "Request with Log ID {0}: {1}".format(self.log_id, msg)
            else:
                return msg

    # Returns the underlying `Exception` (base class) message, which is usually
    # the raw message returned by Fintecture's API. This was previously available
    # in python2 via `error.message`. Unlike `str(error)`, it omits "Request
    # req_..." from the beginning of the string.
    @property
    def user_message(self):
        return self._message

    def __repr__(self):
        return (
            "%s(message=%r, url=%r, http_status=%r, log_id=%r, x_request_id=%r, errors=%r)"
            % (
                self.__class__.__name__,
                self._message,
                self.url,
                self.http_status,
                self.log_id,
                self.x_request_id,
                self.errors,
            )
        )

    def construct_error_object(self):
        if (
            self.json_body is None
            or "errors" not in self.json_body
            or not isinstance(self.json_body["errors"], list)
            or len(self.json_body["errors"]) == 0
        ):
            return None

        return sdk.api_resources.error_object.ErrorObject.construct_from(
            self.json_body, sdk.app_id
        )


class APIError(FintectureError):
    # used when an error occur trying to parse JSON,
    # decoding response data doesn't match documented API scheme, and more
    pass


class APIConnectionError(FintectureError):
    def __init__(
        self,
        message,
        url=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
        should_retry=False,
    ):
        super(APIConnectionError, self).__init__(
            message, url, http_body, http_status, json_body, headers, code
        )
        self.should_retry = should_retry


class FintectureErrorWithParamCode(FintectureError):
    def __repr__(self):
        return (
            "%s(message=%r, url=%r, param=%r, code=%r, http_status=%r, log_id=%r, x_request_id=%r)"
            % (
                self.__class__.__name__,
                self._message,
                self.url,
                self.param,
                self.code,
                self.http_status,
                self.log_id,
                self.x_request_id,
            )
        )


class InvalidRequestError(FintectureErrorWithParamCode):
    def __init__(
        self,
        message,
        param,
        code=None,
        url=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
    ):
        super(InvalidRequestError, self).__init__(
            message, url, http_body, http_status, json_body, headers, code
        )

        # define the name of invalid parameters separate each one by comma
        self.param = param


class AuthenticationError(FintectureError):
    pass


class BadRequestError(FintectureError):
    pass


class AuthorizationError(FintectureError):
    pass


class PermissionError(FintectureError):
    pass


class NotFoundError(FintectureError):
    pass


class TooManyRequestsError(FintectureError):
    pass


class InternalServerError(FintectureError):
    pass


class ServiceUnavailableError(FintectureError):
    pass


class SignatureVerificationError(FintectureError):
    def __init__(
        self,
        message,
        signature_header,
        digest_header,
        request_id_header,
        http_body=None,
    ):
        super(SignatureVerificationError, self).__init__(
            message=message, http_body=http_body
        )
        self.sig_header = signature_header
        self.digest_header = digest_header
        self.request_id_header = request_id_header



import calendar
import datetime
import json
import sys
import time
import warnings

from collections import OrderedDict
from email import utils as email_utils
from urllib.parse import urlparse, urlencode, urlsplit, urlunsplit

# Get reference to SDK module (works for any package name)
import sys

sdk = sys.modules[__name__.split(".")[0]]
from . import crypto
from . import error, http_client, util
from .multipart_data_generator import MultipartDataGenerator
from .fintecture_response import FintectureResponse
from .constants import SIGNED_HEADER_PARAMETER_LIST


def _encode_datetime(dttime):
    if dttime.tzinfo and dttime.tzinfo.utcoffset(dttime) is not None:
        utc_timestamp = calendar.timegm(dttime.utctimetuple())
    else:
        utc_timestamp = time.mktime(dttime.timetuple())

    return int(utc_timestamp)


def _encode_nested_dict(key, data, fmt="%s[%s]"):
    d = OrderedDict()
    for subkey, subvalue in data.items():
        d[fmt % (key, subkey)] = subvalue
    return d


def _api_encode(data):
    for key, value in data.items():
        key = util.utf8(key)
        if value is None:
            continue
        elif hasattr(value, "fintecture_id"):
            yield (key, value.fintecture_id)
        elif isinstance(value, list) or isinstance(value, tuple):
            for i, sv in enumerate(value):
                if isinstance(sv, dict):
                    subdict = _encode_nested_dict("%s[%d]" % (key, i), sv)
                    for k, v in _api_encode(subdict):
                        yield (k, v)
                else:
                    yield ("%s[%d]" % (key, i), util.utf8(sv))
        elif isinstance(value, dict):
            subdict = _encode_nested_dict(key, value)
            for subkey, subvalue in _api_encode(subdict):
                yield (subkey, subvalue)
        elif isinstance(value, datetime.datetime):
            yield (key, _encode_datetime(value))
        else:
            yield (key, util.utf8(value))


def _build_api_url(url, query):
    scheme, netloc, path, base_query, fragment = urlsplit(url)

    if base_query:
        query = "%s&%s" % (base_query, query)

    return urlunsplit((scheme, netloc, path, query, fragment))


class APIRequestor(object):

    def __init__(
        self,
        app_id=None,
        app_secret=None,
        private_key=None,
        client=None,
        api_base=None,
        api_version=None,
    ):
        self.api_base = api_base or APIRequestor.api_base()

        self.app_id = app_id
        self.app_secret = app_secret
        self.private_key = private_key

        self.api_version = api_version or sdk.api_version

        self._default_proxy = None

        verify = sdk.verify_ssl_certs
        proxy = sdk.proxy

        if client:
            self._client = client
        elif sdk.default_http_client:
            self._client = sdk.default_http_client
            if proxy != self._default_proxy:
                warnings.warn(
                    "sdk.proxy was updated after sending a "
                    "request - this is a no-op. To use a different proxy, "
                    "set sdk.default_http_client to a new client "
                    "configured with the proxy."
                )
        else:
            # If the sdk.default_http_client has not been set by the user
            # yet, we'll set it here. This way, we aren't creating a new
            # HttpClient for every request.
            sdk.default_http_client = http_client.new_default_http_client(
                verify_ssl_certs=verify, proxy=proxy
            )
            self._client = sdk.default_http_client
            self._default_proxy = proxy

    @classmethod
    def api_base(cls):
        production_api_base = sdk.production_api_base
        sandbox_api_base = sdk.sandbox_api_base
        env = sdk.env

        if env not in sdk.AVAILABLE_ENVS:
            raise ValueError(
                "Defined environment value is invalid. "
                "Please check that specified environment value is one of %r\n"
                % sdk.AVAILABLE_ENVS
            )
        if env == sdk.ENVIRONMENT_SANDBOX:
            return sandbox_api_base
        elif env == sdk.ENVIRONMENT_PRODUCTION:
            return production_api_base

        return sandbox_api_base

    @classmethod
    def format_app_info(cls, info):
        str = info["name"]
        if info["version"]:
            str += "/%s" % (info["version"],)
        if info["url"]:
            str += " (%s)" % (info["url"],)
        return str

    def request(self, method, url, params=None, headers=None):
        rbody, rcode, rheaders, url, my_app_id = self.request_raw(
            method.lower(), url, params, headers
        )
        resp = self.interpret_response(url, rbody, rcode, rheaders)
        return resp, my_app_id

    def handle_error_response(self, url, rbody, rcode, resp, rheaders):
        try:
            errors = resp["errors"]
        except (KeyError, TypeError):
            raise error.APIError(
                message="Invalid response object from API: %r "
                "(HTTP response code was %d)" % (rbody, rcode),
                url=url,
                http_body=rbody,
                http_status=rcode,
                headers=rheaders,
                json_body=resp,
            )

        err = self.specific_api_error(
            url, rbody, rcode, resp, rheaders, errors
        )

        if err is not None:
            raise err

        raise error.APIError(
            message="Invalid response object from API: %r "
            "(HTTP response code was %d)" % (rbody, rcode),
            url=url,
            http_body=rbody,
            http_status=rcode,
            headers=rheaders,
            json_body=resp,
        )

    def specific_api_error(
        self, url, rbody, rcode, resp, rheaders, error_data
    ):

        status = resp["status"]
        code = resp["code"]

        util.log_info(
            "Fintecture API error received",
            code=code,
            status=status,
            errors=error_data,
        )

        args = [url, rbody, rcode, resp, rheaders, code]

        if rcode == 400 and code == "bad_request":
            bad_errors = [
                {
                    "bad_request": "Invalid parameters or malformed syntax.",
                    "customer_unknown": "Invalid customer_id. "
                    "Use a valid customer_id or authenticate to a bank to continue.",
                    "account_unknown": "Invalid account_id. "
                    "You must specify an account_id as defined by the /accounts API.",
                    "session_id_invalid_or_expired": "The session ID used is either expired or invalid.",
                    "invalid_field": "The value or format of field [field] is incorrect",
                    "mandatory_field_missing": "The mandatory field is missing: [field] has not been defined.",
                    "invalid_debited_account": "Invalid debited_account_id. "
                    "The debited_account_type is set to internal, "
                    "please use an id provider by the accounts API.",
                }
            ]
            message = (
                "Some general information about the error could be one of [%r]. "
                "More details in specific errors." % bad_errors
            )
            args.insert(0, message)
            return error.BadRequestError(*args)

        elif rcode == 401 and code == "unauthorized":
            authorization_errors = [
                {
                    "invalid_token": "The token is either invalid or expired.",
                    "invalid_scopes": "Your app does not have the necessary scopes to access this API.",
                    "invalid_code": "The authorization code is either wrong or expired.",
                    "invalid_app_id": "Invalid app_id.",
                    "invalid_app_url": "Invalid app redirect URL.",
                }
            ]
            message = (
                "Some general information about the error could be one of [%r]. "
                "More details in specific errors." % authorization_errors
            )
            args.insert(0, message)
            return error.AuthorizationError(*args)

        elif rcode == 403:
            args.insert(0, "Some resource could not be accessed.")
            return error.PermissionError(*args)

        elif rcode == 404 and code == "not_found":
            message = (
                "The requested resource could not be found. "
                "The requested resource either does not exist or is temporarily down"
            )
            args.insert(0, message)
            return error.NotFoundError(*args)

        elif rcode == 429 and code == "too_many_requests":
            message = "The user has sent too many requests in a given amount of time."
            args.insert(0, message)
            return error.TooManyRequestsError(*args)

        elif rcode == 500 and code == "internal_server_error":
            message = "An internal error has occurred. If the error persists, please contact our support."
            args.insert(0, message)
            return error.InternalServerError(*args)

        elif rcode == 503 and code == "service_unavailable":
            message = "The provider is currently unavailable. Please try again later."
            args.insert(0, message)
            return error.ServiceUnavailableError(*args)

        else:
            args.insert(
                0, "An unknown error occur. Please, see error details."
            )
            return error.APIError(*args)

    def request_headers(
        self,
        app_id,
        method,
        url,
        access_token=None,
        private_key=None,
        body=None,
        supplied_headers=None,
    ):

        python_version = "{}.{}".format(
            sys.version_info.major, sys.version_info.minor
        )
        # Use SDK introspection for library name and version
        library_name = sdk.__name__ + "-sdk-python"
        user_agent = "%s/%s Python/%s" % (
            library_name,
            sdk.__version__,
            python_version,
        )
        if sdk.app_info:
            user_agent += " " + self.format_app_info(sdk.app_info)

        headers = {
            "User-Agent": user_agent,
            "Accept": "application/json",
        }

        if method == "post":
            if supplied_headers is None:
                headers["Content-Type"] = "application/x-www-form-urlencoded"
            elif supplied_headers.get("Content-Type", False):
                headers["Content-Type"] = supplied_headers.get("Content-Type")

        if self.api_version is not None:
            headers["Fintecture-Version"] = self.api_version

        if app_id:
            headers["app_id"] = app_id

        if access_token:
            headers["Authorization"] = "Bearer {}".format(sdk.access_token)

        if private_key:
            payload = ""
            if isinstance(body, str):
                payload = body
            elif isinstance(body, dict):
                payload = json.dumps(body, separators=(",", ":"))

            parsed_url = urlparse(url)
            path = parsed_url.path
            query = parsed_url.query

            if payload and len(payload) > 0:
                headers["Digest"] = "SHA-256=" + crypto.hash_base64(payload)

            headers["Date"] = email_utils.format_datetime(
                datetime.datetime.now(datetime.timezone.utc)
            )
            headers["X-Date"] = headers["Date"]
            headers["X-Request-ID"] = crypto.generate_uuidv4()
            headers["(request-target)"] = (
                method + " " + path + ("?" + query if query else "")
            )
            headers["Signature"] = crypto.create_signature_header(
                headers, app_id, private_key, SIGNED_HEADER_PARAMETER_LIST
            )
            del headers["(request-target)"]

        return headers

    def request_raw(
        self,
        method,
        url,
        params=None,
        supplied_headers=None,
    ):
        """
        Mechanism for issuing an API call
        """

        if self.app_id:
            my_app_id = self.app_id
        else:
            app_id = sdk.app_id

            my_app_id = app_id

        if my_app_id is None:
            raise error.AuthenticationError(
                "No app_id and/or app_secret provided. (HINT: set your app_id and app_secret using "
                '"sdk.app_id = <APP-ID>" and "sdk.app_secret = <APP-SECRET>"). '
                "You can find your application values in your Fintecture developer console at "
                "https://console.fintecture.com/developers, after registering your account as a platform. "
                "See https://docs.fintecture.com/ for details, or email contact@fintecture.com "
                "if you have any questions."
            )

        if self.private_key:
            my_private_key = self.private_key
        else:
            private_key = sdk.private_key

            my_private_key = private_key

        abs_url = "%s%s" % (self.api_base, url)

        encoded_params = urlencode(list(_api_encode(params or {})))

        # Don't use strict form encoding by changing the square bracket control
        # characters back to their literals. This is fine by the server, and
        # makes these parameter strings easier to read.
        encoded_params = encoded_params.replace("%5B", "[").replace("%5D", "]")

        if method == "get" or method == "delete":
            if params:
                abs_url = _build_api_url(abs_url, encoded_params)
            post_data = None
        elif method == "post" or method == "patch":
            if (
                supplied_headers is not None
                and supplied_headers.get("Content-Type")
                == "multipart/form-data"
            ):
                generator = MultipartDataGenerator()
                generator.add_params(params or {})
                post_data = generator.get_post_data()
                supplied_headers["Content-Type"] = (
                    "multipart/form-data; boundary=%s" % (generator.boundary,)
                )
            elif (
                supplied_headers is not None
                and supplied_headers.get("Content-Type") == "application/json"
            ):
                post_data = json.dumps(params, separators=(",", ":"))
            else:
                post_data = encoded_params
        else:
            raise error.APIConnectionError(
                "Unrecognized HTTP method %r. This may indicate a bug in the "
                "Fintecture bindings. Please contact contact@fintecture.com for "
                "assistance." % (method,)
            )

        headers = self.request_headers(
            my_app_id,
            method,
            abs_url,
            sdk.access_token,
            my_private_key,
            post_data,
        )
        if supplied_headers is not None:
            for key, value in supplied_headers.items():
                headers[key] = value

        util.log_info("Request to Fintecture API", method=method, path=abs_url)
        util.log_debug(
            "Post details",
            headers=headers,
            post_data=encoded_params,
        )

        rcontent, rcode, rheaders = self._client.request_with_retries(
            method, abs_url, headers, post_data
        )

        util.log_info(
            "Fintecture API response", path=abs_url, response_code=rcode
        )
        util.log_debug(
            "API response details",
            headers=rheaders,
            body=rcontent,
        )

        return rcontent, rcode, rheaders, url, my_app_id

    def _should_handle_code_as_error(self, rcode):
        return not 200 <= rcode < 300

    def interpret_response(self, url, rbody, rcode, rheaders):
        try:
            if hasattr(rbody, "decode"):
                rbody = rbody.decode("utf-8")
            resp = FintectureResponse(url, rbody, rcode, rheaders)
        except Exception:
            raise error.APIError(
                message="Invalid response body from API: %s "
                "(HTTP response code was %d)" % (rbody, rcode),
                url=url,
                http_body=rbody,
                http_status=rcode,
                headers=rheaders,
            )

        if self._should_handle_code_as_error(rcode):
            self.handle_error_response(url, rbody, rcode, resp.data, rheaders)

        return resp



import datetime
from collections import OrderedDict

import pytest

from tests.conftest import sdk

VALID_API_METHODS = ("get", "post", "delete")


class GMT1(datetime.tzinfo):
    def utcoffset(self, dt):
        return datetime.timedelta(hours=1)

    def dst(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt):
        return "Europe/Prague"


class APIHeaderMatcher(object):
    EXP_KEYS = [
        "app_id",
        "Fintecture-Version",
        "User-Agent",
        "Accept",
    ]
    METHOD_EXTRA_KEYS = {"post": ["Content-Type"]}

    def __init__(
        self,
        app_id=None,
        extra={},
        request_method=None,
        user_agent=None,
        app_info=None,
        fail_platform_call=False,
    ):
        self.request_method = request_method
        self.app_id = app_id or sdk.app_id
        self.extra = extra
        self.user_agent = user_agent
        self.app_info = app_info
        self.fail_platform_call = fail_platform_call

    def __eq__(self, other):
        return (
            self._keys_match(other)
            and self._auth_match(other)
            and self._user_agent_match(other)
            and self._x_fintecture_ua_contains_app_info(other)
            and self._x_fintecture_ua_handles_failed_platform_function(other)
            and self._extra_match(other)
        )

    def __repr__(self):
        return (
            "APIHeaderMatcher(request_method=%s, app_id=%s, extra=%s, user_agent=%s, app_info=%s, fail_platform_call=%s)"
            % (
                repr(self.request_method),
                repr(self.app_id),
                repr(self.extra),
                repr(self.user_agent),
                repr(self.app_info),
                repr(self.fail_platform_call),
            )
        )

    def _keys_match(self, other):
        expected_keys = list(set(self.EXP_KEYS + list(self.extra.keys())))
        if (
            self.request_method is not None
            and self.request_method in self.METHOD_EXTRA_KEYS
        ):
            expected_keys.extend(self.METHOD_EXTRA_KEYS[self.request_method])
        return sorted(other.keys()) == sorted(expected_keys)

    def _auth_match(self, other):
        return other.get("app_id") == self.app_id

    def _user_agent_match(self, other):
        if self.user_agent is not None:
            return other["User-Agent"] == self.user_agent

        return True

    def _x_fintecture_ua_contains_app_info(self, other):
        # Fintecture doesn't use X-Fintecture-Client-User-Agent header
        return True

    def _x_fintecture_ua_handles_failed_platform_function(self, other):
        # Fintecture doesn't use X-Fintecture-Client-User-Agent header
        return True

    def _extra_match(self, other):
        for k, v in self.extra.items():
            if other[k] != v:
                return False

        return True


class QueryMatcher(object):
    def __init__(self, expected):
        self.expected = sorted(expected)

    def __eq__(self, other):
        query = other.split("?")[-1] if "?" in other else other
        parsed = sdk.util.parse_qsl(query)
        return self.expected == sorted(parsed)

    def __repr__(self):
        return "QueryMatcher(expected=%s)" % (repr(self.expected))


class TestAPIRequestor(object):
    valid_path = "/v1/test"

    ENCODE_INPUTS = {
        "dict": {
            "astring": "bar",
            "anint": 5,
            "adatetime": datetime.datetime(2013, 1, 1, tzinfo=GMT1()),
            "atuple": (1, 2),
            "adict": {"foo": "bar", "boz": 5},
            "alist": ["foo", "bar"],
        },
        "list": [1, "foo", "baz"],
        "string": "boo",
        "unicode": "\u1234",
        "datetime": datetime.datetime(2013, 1, 1, second=1, tzinfo=GMT1()),
    }

    ENCODE_EXPECTATIONS = {
        "dict": [
            ("%s[astring]", "bar"),
            ("%s[anint]", "5"),
            ("%s[adatetime]", "1356994800"),
            ("%s[atuple][0]", "1"),
            ("%s[atuple][1]", "2"),
            ("%s[adict][foo]", "bar"),
            ("%s[adict][boz]", "5"),
            ("%s[alist][0]", "foo"),
            ("%s[alist][1]", "bar"),
        ],
        "list": [("%s[0]", "1"), ("%s[1]", "foo"), ("%s[2]", "baz")],
        "string": [("%s", "boo")],
        "unicode": [("%s", "\u1234")],
        "datetime": [("%s", "1356994801")],
    }

    @pytest.fixture
    def requestor(self):
        return sdk.api_requestor.APIRequestor()

    @pytest.fixture
    def http_client(self, mocker):
        mock_client = mocker.Mock(spec=sdk.http_client.HTTPClient)
        mock_client.request_with_retries.return_value = ("{}", 200, {})
        # Patch the default http client
        mocker.patch.object(sdk, "default_http_client", mock_client)
        return mock_client

    @pytest.fixture
    def mock_response(self, http_client):
        def mock_response(res, code=200, headers={}):
            http_client.request_with_retries.return_value = (
                res,
                code,
                headers,
            )

        return mock_response

    @pytest.fixture
    def check_call(self, http_client):
        def check_call(
            method,
            abs_url=None,
            headers=None,
            post_data=None,
        ):
            if not abs_url:
                abs_url = "%s%s" % (sdk.api_base, self.valid_path)
            if not headers:
                headers = APIHeaderMatcher(request_method=method)

            http_client.request_with_retries.assert_called_with(
                method, abs_url, headers, post_data
            )

        return check_call

    def test_dictionary_list_encoding(self):
        params = {"foo": {"0": {"bar": "bat"}}}
        encoded = list(sdk.api_requestor._api_encode(params))
        key, value = encoded[0]

        assert key == "foo[0][bar]"
        assert value == "bat"

    def test_ordereddict_encoding(self):
        params = {
            "ordered": OrderedDict(
                [
                    ("one", 1),
                    ("two", 2),
                    ("three", 3),
                    ("nested", OrderedDict([("a", "a"), ("b", "b")])),
                ]
            )
        }
        encoded = list(sdk.api_requestor._api_encode(params))

        assert encoded[0][0] == "ordered[one]"
        assert encoded[1][0] == "ordered[two]"
        assert encoded[2][0] == "ordered[three]"
        assert encoded[3][0] == "ordered[nested][a]"
        assert encoded[4][0] == "ordered[nested][b]"

    def test_sets_default_http_client(self):
        assert isinstance(
            sdk.default_http_client,
            sdk.http_client.RequestsClient,
        )

    def test_fails_without_app_id(self, requestor):
        sdk.app_id = None

        with pytest.raises(sdk.error.AuthenticationError):
            requestor.request("get", self.valid_path, {})

    def test_invalid_method(self, requestor):
        with pytest.raises(sdk.error.APIConnectionError):
            requestor.request("foo", "bar")



from ..util import merge_dicts
from ..fintecture_object import FintectureObject


class ErrorObject(FintectureObject):
    def refresh_from(
        self,
        values,
        app_id=None,
        fintecture_version=None,
        last_response=None,
    ):
        # Unlike most other API resources, the API will omit attributes in
        # error objects when they have a null value. We manually set default
        # values here to facilitate generic error handling.
        values = merge_dicts(
            {
                "status": None,
                "code": None,
                "log_id": None,
                "errors": [],
            },
            values,
        )
        return super(ErrorObject, self).refresh_from(
            values,
            app_id,
            fintecture_version,
            last_response,
        )



# Version - must match version in pyproject.toml
__version__ = "2.1.1"

# API resources
from .environments import (  # noqa: F401
    DEFAULT_ENV,
    ENVIRONMENT_PRODUCTION,
    ENVIRONMENT_SANDBOX,
    ENVIRONMENT_TEST,
    AVAILABLE_ENVS,
)

# Import API base URLs from constants (tenant-specific)
from .constants import PRODUCTION_API_BASE, SANDBOX_API_BASE  # noqa: F401

# Configuration variables

app_id = None
app_secret = None
private_key = None
env = DEFAULT_ENV

access_token = None

# API base URLs (imported from constants, can be overridden)
production_api_base = PRODUCTION_API_BASE
sandbox_api_base = SANDBOX_API_BASE

api_version = "v2"
verify_ssl_certs = True
proxy = None
default_http_client = None
app_info = None
enable_telemetry = True
max_network_retries = 0

# Set to either 'debug' or 'info', controls console logging
log = None

# API resources
from .api_resources import *  # noqa

# OAuth
from .oauth import OAuth  # noqa

# Webhooks
from .webhook import Webhook, WebhookSignature  # noqa


# Sets some basic information about the running application that's sent along
# with API requests. Useful for plugin authors to identify their plugin when
# communicating with the API.
#
# Takes a name and optional version and plugin URL.
def set_app_info(name, partner_id=None, url=None, version=None):
    global app_info
    app_info = {
        "name": name,
        "partner_id": partner_id,
        "url": url,
        "version": version,
    }



# File generated from our OpenAPI spec

import base64

# Get reference to SDK module (works for any package name)
import sys

sdk = sys.modules[__name__.split(".")[0]]
from .. import api_requestor, error
from .abstract import APIResource
from .payment import Payment


class PIS(APIResource):
    OBJECT_NAME = "pis"

    @classmethod
    def connect(cls, **params):
        return Payment.connect(**params)

    # noinspection PyPackageRequirements
    @classmethod
    def initiate(cls, provider_id, redirect_uri, **params):
        return Payment.initiate(provider_id, redirect_uri, **params)

    @classmethod
    def oauth(cls, app_id=None, app_secret=None, **params):
        app_id = app_id or sdk.app_id
        app_secret = app_secret or sdk.app_secret

        params.update(
            {
                "app_id": app_id,
                "grant_type": "client_credentials",
                "scope": "PIS",
            }
        )
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": "Basic {}".format(
                base64.b64encode(
                    "{}:{}".format(app_id, app_secret).encode("utf-8")
                ).decode("utf-8")
            ),
        }

        requestor = api_requestor.APIRequestor(app_id, app_secret)
        response, _ = requestor.request(
            "post", "/oauth/accesstoken", params, headers
        )

        return response.data

    @classmethod
    def initiate_refund(cls, **params):
        if not params.get("state", False):
            raise error.InvalidRequestError(
                message="state: A state parameter which will be provided back on redirection.",
                param="state",
            )
        state = params.get("state")
        del params["state"]
        params.update({"headers": {"Content-Type": "application/json"}})
        return cls._static_request(
            "post",
            "/pis/v2/refund?state={}".format(state),
            params=params,
        )

    @classmethod
    def request_to_pay(cls, redirect_uri, **params):
        language = params.get("language", "")
        if not language:
            raise error.InvalidRequestError(
                message="language: its a required parameter that is a code of two letters; ex: fr",
                param="language",
            )
        if len(language) != 2:
            raise error.InvalidRequestError(
                message="language parameter is of two characters of length; ex: fr",
                param="language",
            )
        params.update(
            {
                "headers": {
                    "Content-Type": "application/json",
                    "x-language": language,
                }
            }
        )
        del params["language"]

        with_virtualbeneficiary = params.get("with_virtualbeneficiary", False)
        if with_virtualbeneficiary:
            del params["with_virtualbeneficiary"]

        state = params.get("state", False)
        if state:
            del params["state"]

        url = (
            "/pis/v2/request-to-pay?redirect_uri={url}{virtual}{state}".format(
                url=redirect_uri,
                virtual=(
                    "&with_virtualbeneficiary=true"
                    if with_virtualbeneficiary
                    else ""
                ),
                state="&state={}".format(state) if state else "",
            )
        )

        return cls._static_request(
            "post",
            url,
            params=params,
        )

    @classmethod
    def request_for_payout(cls, redirect_uri, **params):
        if not params.get("state", False):
            raise error.InvalidRequestError(
                message="state: A state parameter which will be provided back on redirection.",
                param="state",
            )
        state = params.get("state")
        del params["state"]

        language = params.get("language", "")
        if not language:
            raise error.InvalidRequestError(
                message="language: its a required parameter that is a code of two letters; ex: fr",
                param="language",
            )
        if len(language) != 2:
            raise error.InvalidRequestError(
                message="language parameter is of two characters of length; ex: fr",
                param="language",
            )
        params.update(
            {
                "headers": {
                    "Content-Type": "application/json",
                    "x-language": language,
                }
            }
        )
        del params["language"]
        return cls._static_request(
            "post",
            "/pis/v2/request-for-payout?redirect_uri={url}&state={state}".format(
                url=redirect_uri,
                state=state,
            ),
            params=params,
        )

    @classmethod
    def settlements(cls, **params):
        settlement_id = params.get("settlement_id", False)
        if settlement_id:
            del params["settlement_id"]
            settlement_id = "/%s" % settlement_id
        else:
            settlement_id = ""

        params.update({"headers": {"Content-Type": "application/json"}})
        return cls._static_request(
            "get",
            "/pis/v2/settlements{settlement}".format(settlement=settlement_id),
            params=params,
        )

    @classmethod
    def class_url(cls):
        return "/v1/pis"



# File generated from our OpenAPI spec

import base64

# Get reference to SDK module (works for any package name)
import sys

sdk = sys.modules[__name__.split(".")[0]]
from .. import error

from .abstract import APIResource
from ..constants import DECOUPLED_MODEL_TYPE


class AIS(APIResource):
    OBJECT_NAME = "ais"

    @classmethod
    def oauth(cls, **params):
        if not params.get("code", False):
            raise error.InvalidRequestError(
                message="code parameter is required for authenticate with oAuth through AIS application",
                param="code",
            )

        params.update(
            {
                "grant_type": "authorization_code",
                "scope": "AIS",
                "headers": {
                    "Content-Type": "application/x-www-form-urlencoded",
                    "Authorization": "Basic {}".format(
                        base64.b64encode(
                            "{}:{}".format(sdk.app_id, sdk.app_secret).encode(
                                "utf-8"
                            )
                        ).decode("utf-8")
                    ),
                },
            }
        )
        return cls._static_request(
            "post",
            "/oauth/accesstoken",
            params=params,
        )

    @classmethod
    def connect(cls, **params):
        if not params.get("redirect_uri", False):
            raise error.InvalidRequestError(
                message="redirect_uri: must correspond to one of the URLs provided when creating an "
                "application on the console.",
                param="redirect_uri",
            )
        if not params.get("state", False):
            raise error.InvalidRequestError(
                message="state: a mandatory state parameter which will be provided back on redirection.",
                param="state",
            )
        return cls._static_request(
            "get",
            "/ais/v2/connect",
            params=params,
        )

    @classmethod
    def authorize(cls, provider_id, redirect_uri, **params):

        if not params.get("state", False):
            raise error.InvalidRequestError(
                message="state: a mandatory state parameter which will be provided back on redirection.",
                param="state",
            )

        model = params.get("model", False)
        # access_token = params.get('access_token', False)

        headers = {}

        if model == DECOUPLED_MODEL_TYPE:
            psu_id = params.get("psu_id", False)
            psu_ip_address = params.get("psu_ip_address", False)
            if not psu_id or not psu_ip_address:
                raise error.InvalidRequestError(
                    message="when model is 'decoupled' the 'x-psu-id' and 'x-psu-ip-address' parameters are required.",
                    param="x-psu-id,x-psu-ip-address",
                )
            headers["x-psu-id"] = psu_id
            headers["x-psu-ip-address"] = psu_ip_address

        params.update(
            {
                "response_type": "code",
                "redirect_uri": redirect_uri,
                # 'access_token': access_token,
                "headers": headers,
            }
        )

        return cls._static_request(
            "get",
            "/ais/v1/provider/{provider}/authorize".format(
                provider=provider_id
            ),
            params=params,
        )

    @classmethod
    def decoupled(cls, provider_id, polling_id, **params):
        headers = {}
        psu_id = params.get("psu_id", False)
        psu_ip_address = params.get("psu_ip_address", False)
        if not psu_id or not psu_ip_address:
            raise error.InvalidRequestError(
                message="when model is 'decoupled' the 'x-psu-id' and 'x-psu-ip-address' parameters are required.",
                param="x-psu-id,x-psu-ip-address",
            )
        headers["x-psu-id"] = psu_id
        headers["x-psu-ip-address"] = psu_ip_address

        params.update({"response_type": "code", "headers": headers})

        return cls._static_request(
            "get",
            "/ais/v1/provider/{provider}/authorize/decoupled/{polling}".format(
                provider=provider_id,
                polling=polling_id,
            ),
            params=params,
        )

    @classmethod
    def class_url(cls):
        return "/v1/ais"



# File generated from our OpenAPI spec

from .abstract import CreateableAPIResource
from .abstract import DeletableAPIResource
from .abstract import ListableAPIResource
from .abstract import SearchableAPIResource
from .abstract import UpdateableAPIResource
from .account import Account
from .account_holder import AccountHolder


class Customer(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    SearchableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "customer"

    @classmethod
    def get_accounts(cls, customer_id, **params):
        return Account.search_by_customer(customer_id)

    @classmethod
    def get_account_holders(cls, customer_id, **params):
        return AccountHolder.search_by_customer(customer_id)

    @classmethod
    def get_account_transactions(cls, customer_id, account_id, **params):
        return Account.search_transactions_by_customer_account(
            customer_id, account_id
        )

    @classmethod
    def delete(cls, customer_id, **params):
        return Account.delete(customer_id, params)

    @classmethod
    def class_url(cls):
        return "/ais/v1/customer"



class RequestMetrics(object):
    def __init__(self, x_request_id, request_duration_ms):
        self.x_request_id = x_request_id
        self.request_duration_ms = request_duration_ms

    def payload(self):
        return {
            "x_request_id": self.x_request_id,
            "request_duration_ms": self.request_duration_ms,
        }



# File generated from our OpenAPI spec

from .. import error

from .abstract import SearchableAPIResource
from .abstract import CreateableAPIResource
from .abstract import DeletableAPIResource
from .abstract import ListableAPIResource
from .abstract import UpdateableAPIResource


class Payment(
    SearchableAPIResource,
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "payment"

    @classmethod
    def search(cls, *args, **kwargs):
        return cls._search(search_url="/pis/v2/payments", *args, **kwargs)

    @classmethod
    def search_auto_paging_iter(cls, *args, **kwargs):
        return cls.search(*args, **kwargs).auto_paging_iter()

    @classmethod
    def connect(cls, **params):
        if not params.get("state", False):
            raise error.InvalidRequestError(
                message="state: A state parameter which will be provided back on redirection.",
                param="state",
            )
        state = params.get("state", "")
        del params["state"]
        with_virtualbeneficiary = params.get("with_virtualbeneficiary", False)
        if with_virtualbeneficiary:
            del params["with_virtualbeneficiary"]
        params.update({"headers": {"Content-Type": "application/json"}})
        return cls._static_request(
            "post",
            "/pis/v2/connect?state={}{}".format(
                state,
                (
                    "&with_virtualbeneficiary=true"
                    if with_virtualbeneficiary
                    else ""
                ),
            ),
            params=params,
        )

    @classmethod
    def initiate(cls, provider_id, redirect_uri, **params):
        if not params.get("state", False):
            raise error.InvalidRequestError(
                message="state: A state parameter which will be provided back on redirection.",
                param="state",
            )
        state = params.get("state")
        del params["state"]
        params.update({"headers": {"Content-Type": "application/json"}})
        return cls._static_request(
            "post",
            "/pis/v2/provider/{provider}/initiate?redirect_uri={url}&state={state}".format(
                provider=provider_id, state=state, url=redirect_uri
            ),
            params=params,
        )

    def refund(self, **params):
        params.update(
            {
                "meta": {"session_id": self.get("data", {}).get("id", {})},
                "headers": {"Content-Type": "application/json"},
            }
        )
        return self._request(
            "post",
            "/pis/v2/refund",
            params=params,
        )

    def update(self, **params):
        session_id = self.get("data", {}).get("id", {})
        if not params.get("status", False):
            raise error.InvalidRequestError(
                message="status: is a parameter for update payment attributes and "
                "only accepts 'payment_cancelled' value",
                param="status",
            )
        status = params.get("status")
        del params["status"]

        params.update(
            {
                "meta": {"status": status},
                "headers": {"Content-Type": "application/json"},
            }
        )
        return self._request(
            "patch",
            "/pis/v2/payments/{}".format(session_id),
            params=params,
        )

    @classmethod
    def class_url(cls):
        return "/pis/v2/payments"



# File generated from our OpenAPI spec

from .. import api_requestor, util
from ..fintecture_object import FintectureObject


class Account(FintectureObject):
    OBJECT_NAME = "account"

    @classmethod
    def _static_request(
        cls,
        method_,
        url_,
        app_id=None,
        fintecture_version=None,
        params=None,
    ):
        """Make a static HTTP request to the Fintecture API"""
        params = None if params is None else params.copy()
        fintecture_version = util.read_special_variable(
            params, "fintecture_version", fintecture_version
        )
        fintecture_app_id = util.read_special_variable(
            params, "app_id", app_id
        )
        fintecture_app_secret = util.read_special_variable(
            params, "app_secret", None
        )
        fintecture_private_key = util.read_special_variable(
            params, "private_key", None
        )
        headers = util.read_special_variable(params, "headers", None)

        requestor = api_requestor.APIRequestor(
            app_id=fintecture_app_id,
            app_secret=fintecture_app_secret,
            private_key=fintecture_private_key,
            api_version=fintecture_version,
        )

        response, my_app_id = requestor.request(method_, url_, params, headers)
        return util.convert_to_fintecture_object(
            response, my_app_id, fintecture_version, params
        )

    @classmethod
    def search_by_customer(cls, customer_id, **params):
        return cls._static_request(
            "get",
            "/ais/v1/customer/{}/accounts".format(customer_id),
            params=params,
        )

    @classmethod
    def search_transactions_by_customer_account(
        cls, customer_id, account_id, **params
    ):
        return cls._static_request(
            "get",
            "/ais/v1/customer/{}/accounts/{}/transactions".format(
                customer_id, account_id
            ),
            params=params,
        )

    @classmethod
    def retrieve_by_customer(cls, customer_id, account_id, **params):
        return cls._static_request(
            "get",
            "/ais/v1/customer/{}/accounts/{}".format(customer_id, account_id),
            params=params,
        )

    @classmethod
    def delete(cls, customer_id, **params):
        return cls._static_request(
            "delete",
            "/ais/v1/customer/{}".format(customer_id),
            params=params,
        )



# -*- coding: utf-8 -*-

from tests.conftest import sdk


class TestFintectureError(object):
    def test_formatting(self):
        err = sdk.error.FintectureError("öre")
        assert str(err) == "öre"

    def test_formatting_with_request_id(self):
        err = sdk.error.FintectureError("öre", headers={"x-request-id": "123"})
        assert str(err) == "Request 123: öre"

    def test_formatting_with_none(self):
        err = sdk.error.FintectureError(None, headers={"x-request-id": "123"})
        assert str(err) == "Request 123: <empty message>"

    def test_formatting_with_message_none_and_request_id_none(self):
        err = sdk.error.FintectureError(None)
        assert str(err) == "<empty message>"

    def test_repr(self):
        err = sdk.error.FintectureError("öre", headers={"x-request-id": "123"})
        assert (
            repr(err)
            == "FintectureError(message='öre', url=None, http_status=None, log_id=None, "
            "x_request_id='123', errors=None)"
        )

    def test_error_object(self):
        err = sdk.error.FintectureError(
            "message",
            json_body={
                "status": "477",
                "code": "invalid_test_code",
                "log_id": "2398-a872-323a-xxxx",
                "errors": [
                    {
                        "code": "some_invalid_code",
                        "title": "Title of test error",
                        "detail": "Details of test error with invalid code",
                    }
                ],
            },
        )
        assert err.error is not None
        assert err.error.code == "invalid_test_code"
        assert err.error.status == "477"
        assert len(err.error.errors) == 1
        assert err.error.errors[0].code == "some_invalid_code"

    def test_error_object_not_dict(self):
        err = sdk.error.FintectureError(
            "message", json_body={"error": "not a dict"}
        )
        assert err.error is None


class TestApiConnectionError(object):
    def test_default_no_retry(self):
        err = sdk.error.APIConnectionError("msg")
        assert err.should_retry is False

        err = sdk.error.APIConnectionError("msg", should_retry=True)
        assert err.should_retry



import time
import base64

from email.utils import parsedate

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.exceptions import UnsupportedAlgorithm

# Import the SDK package dynamically using relative imports
from . import crypto
from . import error, util
from urllib.parse import urlencode
from .api_requestor import _api_encode

# Get reference to the SDK module (works for any package name)
import sys

sdk = sys.modules[__name__.split(".")[0]]


class Webhook(object):
    DEFAULT_TOLERANCE = 300

    @staticmethod
    def construct_event(
        payload,
        digest_header,
        signature_header,
        request_id_header,
        private_key=None,
        tolerance=DEFAULT_TOLERANCE,
        app_id=None,
    ):
        if hasattr(payload, "decode"):
            payload = payload.decode("utf-8")

        pkey = private_key or sdk.private_key
        if pkey is None:
            raise error.AuthenticationError(
                "No private key provided. Set it using: <sdk_name>.private_key = <PRIVATE-KEY>"
            )

        encoded_payload = urlencode(list(_api_encode(payload or {})))

        WebhookSignature.verify_header(
            encoded_payload,
            digest_header,
            signature_header,
            request_id_header,
            pkey,
            tolerance,
        )

        event = util.convert_to_fintecture_object(
            payload, app_id or sdk.app_id
        )

        return event


class WebhookSignature(object):
    EXPECTED_SCHEME = "v2"

    @staticmethod
    def _get_timestamp_and_signatures(header, scheme):
        list_items = [i.split("=", 2) for i in header.split(",")]
        timestamp = int([i[1] for i in list_items if i[0] == "t"][0])
        signatures = [i[1] for i in list_items if i[0] == scheme]
        return timestamp, signatures

    @staticmethod
    def _extract_signature(signature):
        if hasattr(signature, "decode"):
            signature.decode("ascii")
        sign_items = signature.split(
            ","
        )  # 0: keyId, 1: algorithm, 2: headers, 3: signature
        signature_chunk = sign_items[3]
        # now just keep the part after "signature="
        signature_chunk = signature_chunk.replace('"', "")
        signature_value = signature_chunk[
            10:
        ]  # exclude 'signature=' from begin of string
        return signature_value

    @classmethod
    def verify_header(
        cls,
        payload,
        digest_header,
        signature_header,
        request_id_header,
        private_key,
        tolerance=None,
    ):
        try:
            extracted_signature = cls._extract_signature(signature_header)
        except Exception:
            raise error.SignatureVerificationError(
                "Unable to extract signatures from header",
                signature_header,
                digest_header,
                request_id_header,
                payload,
            )

        try:
            bytes_decoded_signature = base64.b64decode(extracted_signature)
        except Exception:
            raise error.SignatureVerificationError(
                "Unable to decode signatures",
                signature_header,
                digest_header,
                request_id_header,
                payload,
            )

        try:
            bytes_private_key = str.encode(private_key)
            private_key_obj = load_pem_private_key(
                data=bytes_private_key,
                password=None,
                backend=default_backend(),
            )
        except ValueError:
            raise error.SignatureVerificationError(
                "the PEM data could not be decrypted or if its structure could not be decoded successfully",
                signature_header,
                digest_header,
                request_id_header,
                payload,
            )

        signature_lines = []
        signature_items = {
            "date": None,
            "digest": None,
            "x-request-id": None,
        }

        try:
            signature = private_key_obj.decrypt(
                ciphertext=bytes_decoded_signature,
                padding=padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA1()),
                    algorithm=hashes.SHA1(),
                    label=None,
                ),
            )

            if hasattr(signature, "decode"):
                signature = signature.decode("utf-8")
                signature_lines = signature.split("\n")
                for sign_item in signature_lines:
                    pair = sign_item.split(": ")
                    signature_items[pair[0]] = pair[1].lstrip()

        except UnsupportedAlgorithm:
            raise ValueError(
                "the serialized key type is not supported by the OpenSSL version cryptography is using"
            )
        except Exception:
            raise error.SignatureVerificationError(
                "Unable to decode signatures",
                signature_header,
                digest_header,
                request_id_header,
                payload,
            )

        if not signature_lines or len(signature_lines) == 0:
            raise error.SignatureVerificationError(
                "No signatures found with expected scheme "
                "%s" % cls.EXPECTED_SCHEME,
                signature_header,
                digest_header,
                request_id_header,
                payload,
            )

        digest = "SHA-256=" + crypto.hash_base64(payload)

        match_digests = util.secure_compare(digest, signature_items["digest"])
        match_request_ids = util.secure_compare(
            request_id_header, signature_items["x-request-id"]
        )

        if match_digests is False or match_request_ids is False:
            raise error.SignatureVerificationError(
                "No signatures found matching the expected signature for payload and request identifier",
                signature_header,
                digest_header,
                request_id_header,
                payload,
            )

        # date_sign = parsedate_tz(signature_items['date'])
        date_sign = parsedate(signature_items["date"])
        timestamp = time.mktime(date_sign)
        time_diff = time.time() - tolerance
        if tolerance and timestamp < time_diff:
            raise error.SignatureVerificationError(
                "Timestamp outside the tolerance (%d) zone (%d)"
                % (tolerance, timestamp),
                signature,
                digest_header,
                request_id_header,
                payload,
            )

        return True



# flake8: noqa

from .api_resource import APIResource
from .singleton_api_resource import (
    SingletonAPIResource,
)

from .createable_api_resource import (
    CreateableAPIResource,
)
from .updateable_api_resource import (
    UpdateableAPIResource,
)
from .deletable_api_resource import (
    DeletableAPIResource,
)
from .listable_api_resource import (
    ListableAPIResource,
)
from .searchable_api_resource import (
    SearchableAPIResource,
)
from .verify_mixin import VerifyMixin

from .custom_method import custom_method

from .test_helpers import (
    test_helpers,
    APIResourceTestHelpers,
)

from .nested_resource_class_methods import (
    nested_resource_class_methods,
)



# -*- coding: utf-8 -*-

import json
import uuid
import base64

from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.exceptions import UnsupportedAlgorithm

from hashlib import sha256


def generate_uuid():
    return str(uuid.uuid4()).replace("-", "")


def generate_uuidv4():
    return str(uuid.uuid4())


def hash_base64(plain_text):
    m = sha256()
    m.update(plain_text.encode("utf-8"))
    b_digest = m.digest()
    b_b64 = base64.b64encode(b_digest)
    plain_b64_hash = b_b64.decode("ascii")
    return plain_b64_hash


def create_signature_header(headers, app_id, private_key, signed_headers):
    signing_string = build_signing_string(headers, signed_headers)
    header_string = build_header_string(headers, signed_headers)

    signature = sign_payload(signing_string, private_key)
    return (
        'keyId="'
        + app_id
        + '",algorithm="rsa-sha256",headers="'
        + header_string
        + '",signature="'
        + signature
        + '"'
    )


def build_signing_string(headers, signed_headers):
    signing_string = ""

    for param in signed_headers:
        if param in headers:
            p = param.lower()
            signing_string = (
                signing_string + "\n" if signing_string else signing_string
            )
            signing_string = signing_string + p + ": " + headers[param]

    return signing_string


def build_header_string(headers, signed_headers):
    header_string = ""

    for param in signed_headers:
        if param in headers:
            p = param.lower()
            header_string = header_string + " " + p if header_string else p

    return header_string


def sign_payload(payload, private_key, algorithm=None):
    if isinstance(payload, dict):
        payload = json.dumps(payload, separators=(",", ":"))

    bytes_payload = str.encode(payload)

    if not algorithm or algorithm == "rsa-sha256":
        try:
            bytes_private_key = str.encode(private_key)
            private_key = load_pem_private_key(
                bytes_private_key, password=None
            )

            signature = private_key.sign(
                bytes_payload, padding.PKCS1v15(), hashes.SHA256()
            )

            base64_bytes = base64.b64encode(signature)
            plain_b64_signature = base64_bytes.decode("ascii")

            return plain_b64_signature

        except ValueError:
            raise ValueError(
                "the PEM data could not be decrypted or "
                "if its structure could not be decoded successfully"
            )
        except TypeError:
            raise ValueError(
                "password was given and the private key was not encrypted; "
                "or the key was encrypted but no password was supplied"
            )
        except UnsupportedAlgorithm:
            raise ValueError(
                "the serialized key type is not supported by the OpenSSL version cryptography is using"
            )

    raise ValueError("invalid signature algorithm")



from .api_resource import APIResource


class CreateableAPIResource(APIResource):
    @classmethod
    def create(cls, app_id=None, fintecture_version=None, **params):
        return cls._static_request(
            "post",
            cls.class_url(),
            app_id,
            fintecture_version,
            params,
        )
